/*
 * file:        main.C
 * purpose:     sample application to show usage of c++ class lib
 *              It additionally shows, how a turning encoder can manipulate
 *              various values. This sample is cluttered by several defines
 *              to determine size of different modules.
 *              Usually an application will be built for a given device, so there's
 *              no need to use processor dependant defines.
 *              But as this sample should be usable for different devices ...
 * project:     c++ framework for avr controllers
 * author:      Reinhard Mantey
 * copyright:   2003-2016 (c) Creative commons - by-nc-sa - some rights reserved
 *              see http://creativecommons.org/licenses/by-nc-sa/3.0/ for details
 */
#define PLACE_GLOBALS_HERE
#include "halConfig.h"
#include "Buzzer.h"
#include "SystemTick.h"
#include "PWM.h"
#include "Led.h"
#include "ADC.h"
#include "PushButton.h"
#include "Encoder.h"
#include "UArt.h"
#include "LCD.h"
#include "UI.h"
#include <avr/io.h>
#include <util/atomic.h>
#include <stdlib.h>

using namespace srd;

#ifdef HAVE_LCD
extern void lcdDisplay(LCD* lcd, uint32_t* secs, int8_t n, int16_t *v);
#endif

#ifdef HAVE_ENCODER
// change application state (work.mode) by turning encoder
class WorkModeHandler : public DefaultEncoderCounterHandler
{
public:
  WorkModeHandler()
   : maxValue(static_cast<int16_t>(WorkMode::MD_Light)) {
    };

  virtual void handleCounter(int8_t delta) {
    work.mode += delta;
    if (work.mode > maxValue) work.mode = 0;
    if (work.mode < 0) work.mode = maxValue;
    };

  int16_t getCounter() const { return static_cast<int16_t>(work.mode); };

private:
  const int16_t maxValue;
  };


// change countdown value with turning encoder
class ConfigCounterHandler : public DefaultEncoderCounterHandler
{
public:
  ConfigCounterHandler()
   : maxValue(30'000) {
    };

  virtual void handleCounter(int8_t delta) {
    work.counter += delta;
    if (work.counter > maxValue) work.counter = 0;
    if (work.counter < 0) work.counter = maxValue;
    };

  int16_t getCounter() const { return work.counter; };

private:
  const int16_t maxValue;
  };


// change pwm duty value with turning encoder
class LightDutyHandler : public DefaultEncoderCounterHandler
{
public:
  LightDutyHandler(int16_t maxValue)
   : maxValue(maxValue) {
    };

  virtual void handleCounter(int8_t delta) {
    counter += delta;
    if (counter > maxValue) counter = 0;
    if (counter < 0) counter = maxValue;
    };

  int16_t getCounter() const { return work.counter; };

private:
  int16_t counter;
  const int16_t maxValue;
  };
#endif


#if defined(HAVE_ADC) || defined(HAVE_SERIAL_COMM)
struct Transfer {
  uint32_t  seconds;
  uint16_t  millis;
  int16_t   adc[8];
  };
#endif


#ifdef HAVE_SERIAL_COMM
class BinaryByteHandler : public ByteHandler
{
public:
  BinaryByteHandler(uint8_t* buf, int16_t bufSize)
   : buf(buf)
   , p(buf)
   , bufSize(bufSize) {
    };
  virtual ~BinaryByteHandler() {};
  virtual void byteReceived(uint8_t byte) {
    if (p < (buf + bufSize - 1)) {
       *p++ = byte;
       }
    else {
       p = buf;
       }
    };
  uint8_t getReceiveCount() const { return receiveCounter; }

private:
  uint8_t* buf;
  uint8_t* p;
  int16_t  bufSize;
  uint8_t  receiveCounter;
  };
#endif


// compiler complains for it ...
void operator delete(void* unused) {
  }
void operator delete(void* unused, unsigned int notUsed) {
  }


/** ***************************************************************************
 *
 *                                  application code ...
 *
 *  ***************************************************************************
 */
int main() {
#ifdef HAVE_SYSTEMTICK
  SystemTick sysTick(1000);                                // internal hwtimer runs at 1kHz
                                                           // so sysTick is 1ms
#endif
#ifdef HAVE_BUZZER
  // buzzer creates an internal hwtimer timer2
  Buzzer<Timer2> buzzer;
#endif
  LED ErrorLed(Pin(PortB, 2));
  LED StatusLed(Pin(PortB, 3));
  PushButton t1(Pin(PortD, 0));
  PushButton t2(Pin(PortD, 1));
#ifndef HAVE_ENCODER
  PushButton t3(Pin(PortD, 2));
  PushButton t4(Pin(PortD, 3));
#else
  Encoder encoder(Pin(PortD, 2), Pin(PortD, 3), PushButton(Pin(PortD, 7)));
#endif
#ifdef HAVE_LCD
  LCD lcd(Pin(PortD, 6), Pin(PortD, 4), Pin(PortD, 5));
#endif
#ifndef ADC_WITH_TIMER
  // if adc does not use the timer, we can create a timer for pwm usage
  // prescale 1 is 62,5 kHz at 16MHz system clock
  Timer0 timer0(TimerMode::FastPWM, PreScale::PS_1, false);

  // timer0 has only one channel an is located at position B
  PWM pwm08_0(timer0, Channel::B, ChannelMode::Toggle);
#endif
#ifdef HAVE_SYSTEMTICK
  // here we use the timer created by sysTick (1kHz in this sample)
  PWM pwm16_A(sysTick.getTimer(), Channel::A, ChannelMode::Toggle);
  PWM pwm16_B(sysTick.getTimer(), Channel::B, ChannelMode::Toggle);
#else
  // here we create a free running 16bit timer with ICR as top
  Timer1 timer1(TimerMode::CTC_ICR, PreScale::PS_1, false);
  PWM pwm16_A(timer1, Channel::A, ChannelMode::Toggle);
  PWM pwm16_B(timer1, Channel::B, ChannelMode::Toggle);
#endif
#ifdef HAVE_SERIAL_COMM
  struct Transfer xfer;
  SerialSender ss(Serial::BR_250k);
  SerialReceiver sr(Serial::BR_250k);
  BinaryByteHandler bbh((uint8_t*)&xfer, sizeof(xfer));
  uint8_t oReceived = 0;
#endif
#ifdef HAVE_ADC
#ifdef ADC_WITH_TIMER
  Timer0 timer0(TimerMode::CTC_OCR, PreScale::PS_1, false);
  timer0.setCompareValue(Channel::B, 40);             // 400kHz
  ADConverter adc(ADConverter::RefVoltage::AVCC, ADConverter::PreScale::PS_2);
#else
  ADConverter adc(ADConverter::RefVoltage::AVCC, ADConverter::PreScale::PS_128);
#endif
#if   defined(__AVR_ATmega16__)
  Pin(PortA, 5).toInput();                                 // connect an unused adc channel to gnd
#elif defined(__AVR_ATmega88__)
  Pin(PortC, 5).toInput();                                 // to speed up and stabilize measurements
#endif
  int8_t numConversions = 0;
#endif
#if defined(HAVE_LCD) || defined(HAVE_ADC) || defined(HAVE_SERIAL_COMM)
  int16_t adcValues[8] = {0};
#endif
  int16_t oTicks = 0;                                      // milli second counter
#if defined(HAVE_SYSTEMTICK) || defined(HAVE_LCD) || defined(HAVE_SERIAL_COMM)
  uint32_t oTime = 0;                                      // second counter
#endif
#ifdef HAVE_SYSTEMTICK
  int16_t t1Top = F_CPU / 1000;                            // 1kHz
#else
  int16_t t1Top = F_CPU / 10'000;                          // 10kHz
#endif
  int16_t duty16_A =   500;
  int16_t duty08_0 =   127;
  bool    changed  = false;
#ifdef HAVE_ENCODER
  // three handler to handle counter value changed by turning encoder
  WorkModeHandler       wmHdr;
  ConfigCounterHandler  ccHdr;
  LightDutyHandler      ldHdr(t1Top);
#else
  int16_t duty16_B = 1'000;
#endif

  /*
   * initializations ...
   */
#ifndef HAVE_SYSTEMTICK
  timer1.setICRValue(t1Top);
#endif
#ifdef HAVE_BUZZER
  buzzer.doBeep(Tones::ToneH4, 600);
#endif
  work.mode = static_cast<int8_t>(WorkMode::MD_Config);
#ifdef HAVE_ENCODER
  encoder.setCounterHandler(&wmHdr);
#endif
#ifndef ADC_WITH_TIMER
  pwm08_0.setDuty(127);
#endif
  pwm16_A.setDuty(99);
  pwm16_B.setDuty(800);
#ifdef HAVE_SERIAL_COMM
  sr.setByteHandler(&bbh);
#endif
#ifdef HAVE_ADC
#ifdef ADC_WITH_TIMER
  adc.setAutoTriggerMode(ADConverter::AutoTrigger::Timer0Compare);
#else
  adc.setAutoTriggerMode(ADConverter::AutoTrigger::FreeRunning);
#endif
#if   defined(__AVR_ATmega16__)                            // mega88 does not have portA,
  adc.enableChannel(Pin(PortA, 0));                        // so we need this hack here
  adc.enableChannel(Pin(PortA, 1));
  adc.enableChannel(Pin(PortA, 2));
  adc.enableChannel(Pin(PortA, 3));
#elif defined(__AVR_ATmega88__)
  adc.enableChannel(Pin(PortC, 0));
  adc.enableChannel(Pin(PortC, 1));
  adc.enableChannel(Pin(PortC, 2));
  adc.enableChannel(Pin(PortC, 3));
#endif
  adc.enableAutoTrigger();
#endif
#ifdef HAVE_LCD
  lcd.clear();
#endif

  /*
   * main loop
   */
  for (;;) {
      // -------------------------------------------------
      // any time
      // -------------------------------------------------
      if (t1.pressedCount() > BUTTON_SAMPLING) {
         t1.reset();
         if (++duty08_0 > 255) duty08_0 = 0;
         }
      if (t2.pressedCount() > BUTTON_SAMPLING) {
         t2.reset();
         if (++duty16_A > t1Top) duty16_A = 0;
         }
#ifndef HAVE_ENCODER
      if (t3.pressedCount() > BUTTON_SAMPLING) {
         t3.reset();
         if (++duty16_B > t1Top) duty16_B = 0;
         }
      if (t4.pressedCount() > BUTTON_SAMPLING) {
         t4.reset();
         ErrorLed.toggle();
         }
#else
      if (encoder.pushButton()->pressedCount() > BUTTON_SAMPLING) {
         encoder.pushButton()->reset();
         ErrorLed.toggle();
         changed = true;
         }
#endif

#ifdef HAVE_ADC
      if (numConversions != adc.conversionCompleted()) {
         numConversions = adc.conversionCompleted();
         for (int i=0; i < 8; ++i) {
             adcValues[i] = adc.getValue(i);
             }
         }
#endif
#ifdef HAVE_SERIAL_COMM
      if (ss.ready2Send()) {
         xfer.millis = oTicks;
         xfer.seconds = oTime;
         for (int i=0; i < 0; ++i) {
             xfer.adc[i] = adcValues[i];
             }
         ss.sendBinary((uint8_t*)&xfer, sizeof(xfer));
         }
      if (bbh.getReceiveCount() != oReceived) {
         oReceived = bbh.getReceiveCount();
         StatusLed.toggle();
         }
#endif

#ifdef HAVE_SYSTEMTICK
      // -------------------------------------------------
      // every millisecond
      // -------------------------------------------------
      if (oTicks != sysTick.getTicks()) {
         ATOMIC_BLOCK(ATOMIC_FORCEON) {
                oTicks = sysTick.getTicks();
                }
#endif
         // put periodic tasks here ...
#ifdef HAVE_BUZZER
         buzzer.countDown();
#endif
#ifdef HAVE_ENCODER
         encoder.handle2Increment();

         if (work.mode == static_cast<int8_t>(WorkMode::MD_Light)) {
            pwm16_B.setDuty(ccHdr.getCounter());
            }
#endif

#ifdef HAVE_SYSTEMTICK
         }                                                 // every millisecond
#endif

      // -------------------------------------------------
      // button debouncing
      // -------------------------------------------------
      if (!(oTicks % IO_CHECK)) {
         t1.checkPressed();
         t2.checkPressed();
#ifdef HAVE_ENCODER
         encoder.pushButton()->checkPressed();
#else
         t3.checkPressed();
         t4.checkPressed();
#endif
         }

#ifdef HAVE_SYSTEMTICK
      // -------------------------------------------------
      // once a second ...
      // -------------------------------------------------
      if (oTime != sysTick.getSeconds()) {
         ATOMIC_BLOCK(ATOMIC_FORCEON) {
                oTime = sysTick.getSeconds();
                }
         // put periodic tasks here ...
         if (work.mode == static_cast<int8_t>(WorkMode::MD_CountDown)) {
            if (work.countDown) {
               if (!(--work.countDown)) {
#ifdef HAVE_BUZZER
                  buzzer.doBeep(Tones::ToneC7, 3000);
#endif
                  work.mode = static_cast<int8_t>(WorkMode::MD_Done);
                  changed = true;
                  }
               }
            }
#endif
#ifdef HAVE_LCD
         lcdDisplay(&lcd, &oTime, 8, adcValues);
         itoa(t1.pressedCount(), work.buf, 10);
         lcd.print(work.buf);
#endif
#ifdef HAVE_SYSTEMTICK
         }                                                 // once a second
#endif

      // -------------------------------------------------
      // asynchronous actions
      // -------------------------------------------------
      if (changed) {
         changed = false;

         // changed application status
         switch (static_cast<WorkMode>(work.mode)) {
           case WorkMode::MD_Config:
#ifdef HAVE_ENCODER
                encoder.setCounterHandler(&ccHdr);
#endif
                break;

           case WorkMode::MD_CountDown:
#ifdef HAVE_ENCODER
                encoder.clearCounterHandler();
#endif
                work.countDown = work.counter;
                break;

           case WorkMode::MD_Light:
#ifdef HAVE_ENCODER
                encoder.setCounterHandler(&ldHdr);
#endif
                break;

           case WorkMode::MD_Done:
#ifdef HAVE_ENCODER
                encoder.setCounterHandler(&wmHdr);
#endif
                work.mode = static_cast<int8_t>(WorkMode::MD_Config);
                break;
           }

#ifdef HAVE_LCD
         lcdDisplay(&lcd, &oTime, 8, adcValues);
#endif
         }
      }
  }
